/*
 * This code implements the RC6-32/20 block cipher.
 *
 * The algorithm is due to Ron Rivest and RSA Labs. This code was written
 * by Martin Hinner <mhi@penguin.cz> in 1999, no copyright is claimed.
 * This code is in the public domain; do with it what you wish.
 *
 * During the AES evaluation period RC6 can be freely used for research
 * and evaluation purposes. For more information see www.rsa.com
 * and AES homepage.
 *
 * RC6 is block cipher using blocks of size 128 bits. To encrypt or
 * decrypt block initialize key structure using rc6keyinit. Then use
 * rc6encrypt/rc6decrypt to encrypt/decrypt block.
 */
#include <stdio.h>
#include "rc6.h"

/*
   Test vector:

   Plaintext:  00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00
   User key:   00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00
   Ciphertext: 8F C3 A5 36 56 B1 F7 78 C1 29 DF 4E 98 48 A4 1E
 */


#define rotl(x, n) (((x) << (n)) | ((x) >> (32-(n))))
#define rotr(x, n) (((x) >> (n)) | ((x) << (32-(n))))


/*
 * Initialize rc6key structure. 'userkey' is user-supplied key of length 'len'.
 */
int
rc6initkey (struct rc6key *key, void *userkey, int len)
{
  int i, j, v, s;
  u_int32_t a, b;
  u_int32_t *l = (u_int32_t *) userkey;

  key->key[0] = RC6_P32;
  for (i = 1; i < RC6_ROUNDS * 2 + 4; i++)
    key->key[i] = key->key[i - 1] + RC6_Q32;

  a = b = i = j = 0;
  v = 3 * (len / 4 > RC6_ROUNDS * 2 + 4 ? len / 4 : RC6_ROUNDS * 2 + 4);

  for (s = 0; s < v; s++)
    {
      a = key->key[i] = rotl (key->key[i] + a + b, 3);
      b = l[j] = rotl (l[j] + a + b, a + b);

      i++;
      i %= RC6_ROUNDS * 2 + 4;

      j++;
      j %= len / 4;
    }

  return 0;
}

/*
 * Encrypts a block of 128 bits (in) using RC6, output goes to 'out'.
 */
int
rc6encrypt (struct rc6key *key, void *in, void *out)
{
  u_int32_t *inl = (u_int32_t *) in;
  u_int32_t *outl = (u_int32_t *) out;
  u_int32_t a, b, c, d, t, u;
  int i;

  a = inl[0];
  b = inl[1];
  c = inl[2];
  d = inl[3];

  b += key->key[0];
  d += key->key[1];

  for (i = 1; i <= RC6_ROUNDS; i++)
    {
      t = rotl ((b * (2 * b + 1)), 5);
      u = rotl ((d * (2 * d + 1)), 5);
      a = rotl (a ^ t, u) + key->key[2 * i];
      c = rotl (c ^ u, t) + key->key[2 * i + 1];

      t = a;

      a = b;
      b = c;
      c = d;
      d = t;
    }
  a += key->key[RC6_ROUNDS * 2 + 2];
  c += key->key[RC6_ROUNDS * 2 + 3];

  outl[0] = a;
  outl[1] = b;
  outl[2] = c;
  outl[3] = d;

  return 0;
}

/*
 * Decrypts a block of 128 bits (in) using RC6, output goes to 'out'.
 */
int
rc6decrypt (struct rc6key *key, void *in, void *out)
{
  u_int32_t *inl = (u_int32_t *) in;
  u_int32_t *outl = (u_int32_t *) out;
  u_int32_t a, b, c, d, t, u;
  int i;

  a = inl[0];
  b = inl[1];
  c = inl[2];
  d = inl[3];

  c -= key->key[RC6_ROUNDS * 2 + 3];
  a -= key->key[RC6_ROUNDS * 2 + 2];

  for (i = RC6_ROUNDS; i > 0; i--)
    {

      t = a;
      u = b;

      a = d;
      d = c;
      b = t;
      c = u;

      u = rotl ((d * (2 * d + 1)), 5);
      t = rotl ((b * (2 * b + 1)), 5);
      c = rotr (c - key->key[2 * i + 1], t) ^ u;
      a = rotr (a - key->key[2 * i], u) ^ t;
    }

  d -= key->key[1];
  b -= key->key[0];

  outl[0] = a;
  outl[1] = b;
  outl[2] = c;
  outl[3] = d;

  return 0;
}
